local _, RS = ...
RS.Events = RS.Events or {}
local Events = RS.Events

-- local references - WoW and Lua
local GetTime = GetTime
local UnitName = UnitName
local UnitIsFriend = UnitIsFriend
local UnitInParty = UnitInParty
local UnitInRaid = UnitInRaid
local IsFalling = IsFalling
local IsSwimming = IsSwimming
local IsFlying = IsFlying

-- local references - Addon
local ActionType = nil
local AttemptReaction = nil
local addonSettings = nil
local playerName = UnitName("player")

function Events:OnLoadMiscellaneous()
	ActionType    = RS.Common.ActionType
	AttemptReaction = RS.Reactions.AttemptReaction
	addonSettings = ReactionsSave
end

----------------------------------------------------------------------------------------------------
-- Data for handling events
----------------------------------------------------------------------------------------------------
-- convert group joining/leaving messages into lua pattern matching strings
local PARTY_STRING_JOINED = "^"..JOINED_PARTY:gsub("%%s", "(%%S+)")
local PARTY_STRING_LEFT   = "^"..LEFT_PARTY:gsub("%%s", "(%%S+)")
local RAID_STRING_JOINED  = "^"..ERR_RAID_MEMBER_ADDED_S:gsub("%%s", "(%%S+)")
local RAID_STRING_LEFT    = "^"..ERR_RAID_MEMBER_REMOVED_S:gsub("%%s", "(%%S+)")
-- other localized text
local ERR_DUEL_REQUESTED  = ERR_DUEL_REQUESTED

----------------------------------------------------------------------------------------------------
-- Handle events
----------------------------------------------------------------------------------------------------
--------------------------------------------------
-- watch health
--------------------------------------------------
function Events:UNIT_HEALTH(unit)
	if unit == "player" then
		local percent = floor(UnitHealth("player") / UnitHealthMax("player") * 100)
		if self.onLowHealth then
			if percent < self.onLowHealth then
				self.onLowHealth = percent
				AttemptReaction("LowHealthLower", 0, nil, playerName, nil, ActionType.HIT)
			elseif percent >= addonSettings.lowHealthEnd then
				self.onLowHealth = nil
				AttemptReaction("LowHealthEnd", 0, nil, playerName, nil, ActionType.HIT)
			end
		elseif percent <= addonSettings.lowHealthBegin then
			self.onLowHealth = percent
			AttemptReaction("LowHealthBegin", 0, nil, playerName, nil, ActionType.HIT)
		end
	end
end

--------------------------------------------------
-- watch mana/rage
--------------------------------------------------
function Events:UNIT_POWER_UPDATE(unit, powerType)
	if powerType == "MANA" then
		if unit == "player" then
			local percent = floor(UnitPower("player", 0) / UnitPowerMax("player", 0) * 100)
			if self.onLowMana then
				if percent < self.onLowMana then
					self.onLowMana = percent
					AttemptReaction("LowManaLower", 0, nil, playerName, nil, ActionType.HIT)
				elseif percent >= addonSettings.lowManaEnd then
					self.onLowMana = nil
					AttemptReaction("LowManaEnd", 0, nil, playerName, nil, ActionType.HIT)
				end
			elseif percent <= addonSettings.lowManaBegin then
				self.onLowMana = percent
				AttemptReaction("LowManaBegin", 0, nil, playerName, nil, ActionType.HIT)
			end
		end
	elseif powerType == "RAGE" then
		if UnitPower(unit, 1) == 100 then
			AttemptReaction("FullRage", 0, nil, UnitName(unit), nil, ActionType.HIT)
		end
	end
end

--------------------------------------------------
-- player changes target
--------------------------------------------------
function Events:PLAYER_TARGET_CHANGED()
	local targetName = UnitName("target")
	if CheckInteractDistance("target", 1) then
		if UnitAffectingCombat("player") then -- in combat
			if UnitIsFriend("player", "target") then
				AttemptReaction((UnitIsDeadOrGhost("target") and "CombatTargetDeadFriend" or "CombatTargetFriend"), 0, nil, targetName, nil, ActionType.HIT)
			else
				AttemptReaction((UnitIsDeadOrGhost("target") and "CombatTargetDeadEnemy" or "CombatTargetEnemy"), 0, nil, targetName, nil, ActionType.HIT)
			end
		else -- not in combat
			if UnitIsFriend("player", "target") then
				AttemptReaction((UnitIsDeadOrGhost("target") and "NewTargetDeadFriend" or "NewTargetFriend"), 0, nil, targetName, nil, ActionType.HIT)
			else
				AttemptReaction((UnitIsDeadOrGhost("target") and "NewTargetDeadEnemy" or "NewTargetEnemy"), 0, nil, targetName, nil, ActionType.HIT)
			end
		end
	end
end

--------------------------------------------------
-- combat starts or stops
--------------------------------------------------
local lastFightStart = nil

function Events:PLAYER_REGEN_DISABLED()
	lastFightStart = GetTime()
	AttemptReaction("CombatBegin", 0, nil, playerName, nil, ActionType.HIT)
end

function Events:PLAYER_REGEN_ENABLED()
	if lastFightStart and GetTime() - lastFightStart >= addonSettings.fightLength then
		addonSettings.fightCounter = addonSettings.fightCounter + 1
	end
	AttemptReaction("CombatEnd", 0, nil, playerName, nil, ActionType.HIT)
end

--------------------------------------------------
-- watch equipment durability
--------------------------------------------------
function Events:UPDATE_INVENTORY_DURABILITY()
	local lowest = RS.Common.GetLowestEquipmentDurability()
	-- only use each event once until the durability is raised above the trigger amounts
	if lowest then
		if lowest == 0 then
			if self.lowestDurability > 0 then
				AttemptReaction("DurabilityBroken", 0, nil, playerName, nil, ActionType.HIT)
			end
		elseif self.lowestDurability > addonSettings.lowDurability then
			if lowest <= addonSettings.lowDurability then
				AttemptReaction("DurabilityLow", 0, nil, playerName, nil, ActionType.HIT)
			end
		end
		self.lowestDurability = lowest
	end
end

--------------------------------------------------
-- loot is opened or closed
--------------------------------------------------
function Events:LOOT_OPENED(autoloot)
	self.autoLootOrTime = autoloot and 1 or GetTime()
	if self.autoLootOrTime == 1 then
		AttemptReaction("LootAutoloot", 0, nil, playerName, nil, ActionType.HIT)
	else
		AttemptReaction("LootOpen", 0, nil, playerName, nil, ActionType.HIT)
	end
end

function Events:LOOT_CLOSED()
	-- sometimes LOOT_CLOSED can come first with empty bags, so must check if autoLootOrTime
	-- was actually set first
	if self.autoLootOrTime and self.autoLootOrTime ~= 1 then
		-- closing the loot very fast triggers the CloseFast since it makes sense for messages for
		-- such speedy looting to be different! It's always possible to set a Spell/Event reaction
		-- on LootCloseFast to use LootClose if this isn't wanted.
		if GetTime() - self.autoLootOrTime <= 1 then
			AttemptReaction("LootCloseFast", 0, nil, playerName, nil, ActionType.HIT)
		else
			AttemptReaction("LootClose", 0, nil, playerName, nil, ActionType.HIT)
		end
	end
	self.autoLootOrTime = nil
end

--------------------------------------------------
-- player quits or logs out
--------------------------------------------------
function Events:PLAYER_CAMPING()
	AttemptReaction(IsResting() and "LogoutInstant" or "LogoutStart", 0, nil, playerName, nil, ActionType.HIT)
end

function Events:PLAYER_QUITING()
	AttemptReaction(IsResting() and "LogoutInstant" or "LogoutStart", 0, nil, playerName, nil, ActionType.HIT)
end

--------------------------------------------------
-- summon accept before/after, cancel
--------------------------------------------------
-- To be able to use a reaction after taking the summon, the GetTime() of accepting it is saved.
-- For short non-loading summons, if a ZONE_CHANGED event happens in less than a second of that time
-- then SummonAcceptAfter is used. For longer summons, if the PLAYER_LEAVING_WORLD event happens in
-- less than a second, then acceptSummonTime is set to 1. When PLAYER_ENTERING_WORLD happens, if it
-- is set to 1 then SummonAcceptAfter is used.
local originalConfirmSummon = C_SummonInfo.ConfirmSummon
C_SummonInfo.ConfirmSummon = function()
	if not UnitAffectingCombat("player") then
		AttemptReaction("SummonAcceptBefore", 0, nil, playerName, C_SummonInfo.GetSummonConfirmSummoner(), ActionType.HIT)
		Events.acceptSummonTime = GetTime()
	end
	originalConfirmSummon()
end

hooksecurefunc(C_SummonInfo, "CancelSummon", function()
	if C_SummonInfo.GetSummonConfirmTimeLeft() > 0 then
		AttemptReaction("SummonCancel", 0, nil, playerName, C_SummonInfo.GetSummonConfirmSummoner(), ActionType.HIT)
	end
end)

--------------------------------------------------
-- entering a new zone or instance
-- before and after being summoned
--------------------------------------------------
function Events:ZONE_CHANGED_NEW_AREA()
	if GetTime() - self.acceptSummonTime <= 5 then
		self.acceptSummonTime = 0
		AttemptReaction("SummonAcceptAfter", 0, nil, playerName, C_SummonInfo.GetSummonConfirmSummoner(), ActionType.HIT)
	end
	local zone = GetRealZoneText()
	if zone and zone ~= "" and zone ~= self.lastZone and AttemptReaction("ChangedZone", 0, nil, playerName, nil, ActionType.HIT) then
		self.lastZone = zone
	end
end

local function CheckChangedZone(events)
	if GetTime() - events.acceptSummonTime <= 2 then
		events.acceptSummonTime = 0
		AttemptReaction("SummonAcceptAfter", 0, nil, playerName, C_SummonInfo.GetSummonConfirmSummoner(), ActionType.HIT)
	end
	local subzone = GetSubZoneText()
	if subzone and subzone ~= "" and subzone ~= events.lastSubzone and subzone ~= GetRealZoneText and AttemptReaction("ChangedSubzone", 0, nil, playerName, nil, ActionType.HIT) then
		events.lastSubzone = subzone
	end
end

function Events:ZONE_CHANGED() CheckChangedZone(self) end
function Events:ZONE_CHANGED_INDOORS() CheckChangedZone(self) end

function Events:PLAYER_LEAVING_WORLD()
	if GetTime() - self.acceptSummonTime <= 1 then
		self.acceptSummonTime = 1
	end
end

function Events:PLAYER_ENTERING_WORLD()
	if self.acceptSummonTime == 1 then
		self.acceptSummonTime = 0
		AttemptReaction("SummonAcceptAfter", 0, nil, playerName, C_SummonInfo.GetSummonConfirmSummoner(), ActionType.HIT)
	end
end

--------------------------------------------------
-- pending party member changes
-- only needed depending on order of events
--------------------------------------------------
function Events:GROUP_ROSTER_UPDATE()
	if not self.inGroup then
		if not UnitIsGroupLeader("player") and IsInGroup() then
			AttemptReaction("GroupJoin", 0, nil, playerName, nil, ActionType.HIT)
		end
		addonSettings.groupCounter = (addonSettings.groupCounter or 0) + 1
		self.inGroup = true
	elseif not IsInGroup() then -- left the group
		self.inGroup = false
		-- clean up tracked auras in case any were left behind
		for k,v in pairs(self.trackedAuras) do
			if k ~= playerName then
				self.trackedAuras[k] = nil
			end
		end
		self.pendingGroupJoinName = nil
		self.pendingGroupLeaveName = nil
		AttemptReaction("GroupLeave", 0, nil, playerName, nil, ActionType.HIT)
		return
	end

	if self.pendingGroupJoinName and (UnitInParty(self.pendingGroupJoinName) or UnitInRaid(self.pendingGroupJoinName)) then
		AttemptReaction("GroupJoin", 0, nil, self.pendingGroupJoinName, nil, ActionType.HIT)
		self.pendingGroupJoinName = nil
	elseif self.pendingGroupLeaveName and not UnitInParty(self.pendingGroupLeaveName) and not UnitInRaid(self.pendingGroupLeaveName) then
		AttemptReaction("GroupLeave", 0, nil, self.pendingGroupLeaveName, nil, ActionType.HIT)
		self.inGroup = IsInGroup()
		self.pendingGroupLeaveName = nil
	end
end

--------------------------------------------------
-- entering/exiting a resting area (inn/capital)
--------------------------------------------------
function Events:PLAYER_UPDATE_RESTING()
	AttemptReaction(IsResting() and "RestingBegin" or "RestingEnd", 0, nil, playerName, nil, ActionType.HIT)
end

--------------------------------------------------
-- opening/closing a merchant window
--------------------------------------------------
function Events:MERCHANT_SHOW()
	self.merchantName = UnitName("npc")
	AttemptReaction("MerchantOpen", 0, nil, playerName, self.merchantName, ActionType.HIT)
end

function Events:MERCHANT_CLOSED()
	if self.merchantName then
		AttemptReaction("MerchantClose", 0, nil, playerName, self.merchantName, ActionType.HIT)
		self.merchantName = nil
	end
end

--------------------------------------------------
-- deleting an item
--------------------------------------------------
function Events:DELETE_ITEM_CONFIRM(itemName)
	local cursorType, _, link = GetCursorInfo()
	if cursorType == "item" then
		RS.Variables.itemName = itemName
		RS.Variables.itemLink = link
	end
end

hooksecurefunc("DeleteCursorItem", function()
	AttemptReaction("DestroyItem", 0, nil, playerName, nil, ActionType.HIT)
end)

--------------------------------------------------
-- jump
--------------------------------------------------
hooksecurefunc("JumpOrAscendStart", function()
	if not IsFalling() and not Events:IsPlayerIncapacitated() and not IsSwimming() and not UnitOnTaxi("player") and not IsFlying() then
		AttemptReaction("Jump", 0, nil, playerName, nil, ActionType.HIT)
	end
end)

--------------------------------------------------
-- focus friend/enemy
--------------------------------------------------
hooksecurefunc("FocusUnit", function()
	local name = UnitName("focus")
	if name and name ~= "Unknown" then
		AttemptReaction(UnitIsFriend("player", "focus") and "FocusFriend" or "FocusEnemy", 0, nil, name, nil, ActionType.HIT)
	end
end)

--------------------------------------------------
-- backpedal
--------------------------------------------------
-- hooksecurefunc("MoveBackwardStart", function()
-- 	if not IsFalling() and not IsPlayerIncapacitated() and not IsSwimming() and not UnitOnTaxi("player") then -- and not IsFlying()
-- 		AttemptReaction("Backpedal", 0, nil, playerName, nil, ActionType.HIT)
-- 	end
-- end)

--------------------------------------------------
-- minimap ping
--------------------------------------------------
function Events:MINIMAP_PING(unitId, x, y)
	if abs(x) <= .5 and abs(y) <= .5 then
		local V = RS.Variables
		V.eventDirection = ""
		if     y <= -0.15 then V.eventDirection = "south"
		elseif y >= 0.15  then V.eventDirection = "north"
		end
		if     x <= -0.15 then V.eventDirection = V.eventDirection .. "west"
		elseif x >= 0.15  then V.eventDirection = V.eventDirection .. "east"
		end
		if V.eventDirection ~= "" then
			V.eventDirection = V.eventDirection
		else
			V.eventDirection = "nearby"
		end
		AttemptReaction("MinimapPing", 0, nil, UnitName(unitId), nil, ActionType.HIT)
	end
end

--------------------------------------------------
-- opening/closing bank and guild bank
--------------------------------------------------
function Events:BANKFRAME_OPENED()
	AttemptReaction("BankOpen", 0, nil, playerName, UnitName("npc"), ActionType.HIT)
end

function Events:BANKFRAME_CLOSED()
	AttemptReaction("BankClose", 0, nil, playerName, UnitName("npc"), ActionType.HIT)
end

--------------------------------------------------
-- opening/closing mailbox
--------------------------------------------------
function Events:MAIL_SHOW()
	AttemptReaction("MailboxOpen", 0, nil, playerName, nil, ActionType.HIT)
	self.mailboxOpened = true
end

function Events:MAIL_CLOSED()
	if self.mailboxOpened then
		self.mailboxOpened = nil
		AttemptReaction("MailboxClose", 0, nil, playerName, nil, ActionType.HIT)
	end
end

--------------------------------------------------
-- accepting or completing a quest
--------------------------------------------------
function Events:QUEST_ACCEPTED()
	-- UntiName("npc") is nil when this happens
	AttemptReaction("QuestAccept", 0, nil, playerName, UnitName("target"), ActionType.HIT)
end

function Events:QUEST_TURNED_IN()
	AttemptReaction("QuestComplete", 0, nil, playerName, UnitName("npc"), ActionType.HIT)
end

--------------------------------------------------
-- duel request by someone else
--------------------------------------------------
function Events:DUEL_REQUESTED(opponentName)
	AttemptReaction("DuelRequest", 0, nil, opponentName, nil, ActionType.HIT)
end

--------------------------------------------------
-- trading - more in UI_INFO_MESSAGE
--------------------------------------------------
function Events:TRADE_SHOW()
	self.tradePartner = UnitName("npc")
	AttemptReaction("TradeOpen", 0, nil, playerName, self.tradePartner, ActionType.HIT)
end

--------------------------------------------------
-- opening/closing auction house
--------------------------------------------------
function Events:AUCTION_HOUSE_SHOW()
	self.auctioneerName = UnitName("npc")
	AttemptReaction("AuctionHouseOpen", 0, nil, playerName, self.auctioneerName, ActionType.HIT)
end

function Events:AUCTION_HOUSE_CLOSED()
	if self.auctioneerName then
		AttemptReaction("AuctionHouseClose", 0, nil, playerName, self.auctioneerName, ActionType.HIT)
		self.auctioneerName = nil
	end
end

--------------------------------------------------
-- ready checks
--------------------------------------------------
function Events:READY_CHECK(requesterName)
	AttemptReaction("ReadyCheckBegin", 0, nil, requesterName, nil, ActionType.HIT)
end

function Events:READY_CHECK_CONFIRM(unitId, isReady)
	local name = UnitName(unitId)
	AttemptReaction(isReady and "ReadyCheckIsReady" or "ReadyCheckNotReady", 0, nil, name, nil, ActionType.HIT)
end

--------------------------------------------------
-- start/stop following another player
--------------------------------------------------
function Events:AUTOFOLLOW_BEGIN()
	self.followingPlayer = UnitName("target")
	AttemptReaction("FollowBegin", 0, nil, playerName, self.followingPlayer, ActionType.HIT)
end

function Events:AUTOFOLLOW_END()
	AttemptReaction("FollowEnd", 0, nil, playerName, self.followingPlayer, ActionType.HIT)
	self.followingPlayer = nil
end

--------------------------------------------------
-- repairing
--------------------------------------------------
hooksecurefunc("RepairAllItems", function(guildBankRepair)
	local cost, canRepair = GetRepairAllCost()
	if canRepair and cost <= GetMoney() then
		AttemptReaction("RepairItems", 0, nil, playerName, UnitName("npc"), ActionType.HIT)
	end
end)

--------------------------------------------------
-- leveling up
--------------------------------------------------
function Events:PLAYER_LEVEL_UP(level)
	RS.Variables.playerLevel = level
	AttemptReaction("LevelUp", 0, nil, playerName, nil, ActionType.HIT)
end

--------------------------------------------------
-- reading a text object
--------------------------------------------------
ItemTextFrame:HookScript("OnShow", function()
	local title = ItemTextTitleText and ItemTextTitleText:GetText()
	if title then
		RS.Variables.itemName = title
		local itemName, found
		for bag=0,4 do
			for slot=1,GetContainerNumSlots(bag) do
				link = GetContainerItemLink(bag, slot)
				if link then
					itemName = link:match("%[(.-)]")
					if itemName and itemName == title then
						RS.Variables.itemLink = link
						found = true
						break
					end
				end
			end
			if found then
				break
			end
		end

		if not found then
			RS.Variables.itemLink = nil
		end

		AttemptReaction("ReadTextItem", 0, nil, nil, nil, ActionType.HIT)
	end
end)

--------------------------------------------------
-- resurrecting
--------------------------------------------------
function Events:PLAYER_DEAD()
	self.isDead = true
end

-- also happens when really logging in (not reloadui) even if you're dead
function Events:PLAYER_ALIVE()
	if not self.hasLoggedIn then
		self.isRealLogin = true
	elseif self.isDead and not UnitIsDeadOrGhost("player") then
		self.isDead = false
		AttemptReaction("Resurrected", 0, nil, playerName, nil, ActionType.HIT)
	end
end

function Events:PLAYER_UNGHOST()
	if self.isDead then
		self.isDead = false
		AttemptReaction("Resurrected", 0, nil, playerName, nil, ActionType.HIT)
	end
end

--------------------------------------------------
-- messages
--------------------------------------------------
function Events:UI_INFO_MESSAGE(_, message)
	-- check for trade messages
	if message == ERR_TRADE_COMPLETE then
		if self.tradePartner then
			AttemptReaction("TradeAccept", 0, nil, playerName, self.tradePartner, ActionType.HIT)
			self.tradePartner = nil
		end
	elseif message == ERR_TRADE_CANCELLED then
		if self.tradePartner then
			AttemptReaction("TradeCancel", 0, nil, playerName, self.tradePartner, ActionType.HIT)
			self.tradePartner = nil
		end
	end

	self:CheckChatTriggers("UI_INFO_MESSAGE", message)
end

function Events:CHAT_MSG_SYSTEM(message)
	self:CheckChatTriggers("CHAT_MSG_SYSTEM", message)

	-- duel request by player - done here because the message is shown only when a request is
	-- successfully made while the Duel cast "success" can be spammed and not work.
	if message == ERR_DUEL_REQUESTED then
		AttemptReaction("DuelRequest", 0, nil, playerName, self.playerCastingTarget, ActionType.HIT)
		return
	end

	-- group changes - GROUP_ROSTER_UPDATE handles when you join or leave the group because
	-- these messages are too unreliable for you because the group type won't be known
	local player

	-- someone joins the group
	player = message:match(PARTY_STRING_JOINED)
	player = player or message:match(RAID_STRING_JOINED)
	if player then
		-- sometimes the message comes before they're in the group - if that happens, then set
		-- the group pending name and wait for the group changing event
		if UnitInParty(player) or UnitInRaid(player) then
			AttemptReaction("GroupJoin", 0, nil, player, nil, ActionType.HIT)
		else
			self.pendingGroupJoinName = player
		end
		return
	end
	-- someone leaves the group
	player = message:match(PARTY_STRING_LEFT)
	player = player or message:match(RAID_STRING_LEFT)
	if player then
		-- sometimes the message comes before they're out of the group - if that happens, then set
		-- the group pending name and wait for the group changing event
		if not UnitInParty(player) and not UnitInRaid(player) then
			AttemptReaction("GroupLeave", 0, nil, player, nil, ActionType.HIT)
		else
			self.pendingGroupLeaveName = player
		end
		return
	end
end

function Events:CHAT_MSG_IGNORED(name)
		RS.Queue:RemoveFutureWhispers(name)
		-- The message for being ignored is just their name, so make it into something to match
		self:CheckChatTriggers("CHAT_MSG_IGNORED", string.format(CHAT_IGNORED, name), name)
end

-- create generic handlers for all the other chat type events
for _,chatType in ipairs({"CHAT_MSG_INSTANCE_CHAT", "CHAT_MSG_INSTANCE_CHAT_LEADER", "CHAT_MSG_CHANNEL",
"CHAT_MSG_EMOTE", "CHAT_MSG_GUILD", "CHAT_MSG_IGNORED", "CHAT_MSG_LOOT", "CHAT_MSG_MONEY",
"CHAT_MSG_MONSTER_EMOTE", "CHAT_MSG_MONSTER_SAY", "CHAT_MSG_MONSTER_YELL", "CHAT_MSG_OFFICER",
"CHAT_MSG_OPENING", "CHAT_MSG_PARTY", "CHAT_MSG_PARTY_LEADER", "CHAT_MSG_RAID", "CHAT_MSG_RAID_LEADER",
"CHAT_MSG_RAID_WARNING", "CHAT_MSG_SAY", "CHAT_MSG_SYSTEM", "CHAT_MSG_TEXT_EMOTE", "CHAT_MSG_TRADESKILLS",
"CHAT_MSG_WHISPER", "CHAT_MSG_YELL", "UI_ERROR_MESSAGE", "UI_INFO_MESSAGE"}) do
	if not Events[chatType] then
		Events[chatType] = function(self, arg1, arg2, arg3, arg4)
			self:CheckChatTriggers(chatType, arg1, arg2, arg3, arg4)
		end
	end
end
