local _, RS = ...
RS.Events = RS.Events or {}
local Events = RS.Events

-- local references - WoW and Lua
local GetTime = GetTime
local IsFalling = IsFalling
local IsSwimming = IsSwimming
local table_remove = table.remove

-- local references - Addon
local ActionType = nil
local AttemptReaction = nil
local playerName = UnitName("player")

function Events:OnLoadTracking()
	ActionType = RS.Common.ActionType
	AttemptReaction = RS.Reactions.AttemptReaction
end

----------------------------------------------------------------------------------------------------
-- helper functions
----------------------------------------------------------------------------------------------------
local nextAutomaticTime = 9999 -- real time set up when addon loads

-- set the time left until the next automatic event happens
function Events:SetNextAutomaticTime()
	nextAutomaticTime = random(ReactionsSave.automaticMinimum*60, ReactionsSave.automaticMaximum*60)
end

----------------------------------------------------------------------------------------------------
-- tracking spell successes/failures, movement, and more
----------------------------------------------------------------------------------------------------
-- miscellaneous tracking
local isSwimming           = nil -- if the player is currently swimming
local isUnderwater         = nil -- if currently underwater (only if not able to breathe)
local nextSwimmingCheck    = 1   -- time left before the next swimming check
local fallingTime           = 0   -- how long the player has been falling
local nextLostControlCheck = nil -- time left before the next check to see if control is still lost
local flightVerified       = nil -- if the taxi flight has been verified to start (it won't start if they pick their current location)
local flightLocationName   = nil -- the destination of the current taxi flight, to be able to set <target_name> for landing
local nextFlightCheck      = nil -- time left before the next taxi flight check

-- spell tracking
-- when a spell cast completes, it's not immediately known if it was a success. The failure messages
-- come soon after the finished event, so a time very near in the future (about 1/2 a second) is set
-- to allow the parsing of the other possible messages. If a message is found, then the tracking
-- variables about it are erased and a miss action occurs. If one isn't found, then the spell is
-- assumed to be successful.
Events.trackedAuras           = {}  -- table of tracked "limitAura" auras for group members and yourself
Events.playerCastingSpellName = nil -- the current watched spell name being casted by the player
Events.playerCastingSpellId   = nil -- the current watched spell ID being casted by the player -- no spell IDs in Classic
Events.playerCastingTarget    = nil -- the name of the target that's currently being casted on by the player
Events.castChecks             = {}  -- tables like: {[1]=GetTime(), [2]=caster, [3]=target, [4]=spell name, [5]=spell ID}
Events.nextCastCheck          = nil -- castChecks[1][1] or nil - set here just for faster checking in OnUpdate

-- return true if a certain spell from a caster is still on the list
function Events:HasCastCheck(caster, spellId)
	local castChecks = self.castChecks
	local info
	for i=1,#castChecks do
		info = castChecks[i]
		if info[5] == spellId and info[2] == caster then
			return true
		end
	end
	return nil
end

-- add a "successfully" cast spell to the list - if it's still on the list in about half a second
-- then it's assumed to have worked
function Events:AddCastCheck(caster, target, spell, id)
	local castChecks = self.castChecks
	local check = {GetTime()+.5, caster, target, spell, id}
	if not castChecks[1] then
		Events.nextCastCheck = check[1]
		castChecks[1] = check
	else
		castChecks[#castChecks+1] = check
	end
end

-- remove a spell from a list, optionally trying a reaction - used when seeing a miss event or an
-- event that did damage/healed/something
function Events:RemoveCastCheck(caster, spell, useReaction)
	local castChecks = self.castChecks
	local info
	for i=1,#castChecks do
		info = castChecks[i]
		if info[2] == caster and info[4] == spell then
			if useReaction then
				AttemptReaction(info[4], info[5], nil, info[2], info[3], ActionType.HIT)
			end
			table_remove(castChecks, i)
			self.nextCastCheck = castChecks[1] and castChecks[1][1] or nil
			return
		end
	end
end

-- use the first spell on the list and possibly more if their times are up
function Events:UseCastChecks()
	local castChecks = self.castChecks
	local info
	repeat
		info = castChecks[1]
		AttemptReaction(info[4], info[5], nil, info[2], info[3], ActionType.HIT)
		table_remove(castChecks, 1)
	until castChecks[1] == nil or GetTime() < castChecks[1][1]
	self.nextCastCheck = castChecks[1] and castChecks[1][1] or nil
end

local elapsedTime = 0
Events.eventFrame:SetScript("OnUpdate", function(frame, elapsed)
	elapsedTime = elapsedTime + elapsed

	-- check falling
	if IsFalling() then
		fallingTime = fallingTime + elapsed
		if fallingTime > 3 then
			fallingTime = 0
			AttemptReaction("Falling", 0, nil, playerName, nil, ActionType.HIT)
		end
	else
		fallingTime = 0
	end

	if elapsedTime < .15 then
		return
	end

	-- check for pending spell cast results - if it's still on the cast check list when its time
	-- passes then it can be assumed that it worked - more miss or hit events would have removed it
	local Events = Events
	if Events.nextCastCheck and GetTime() > Events.nextCastCheck then
		Events:UseCastChecks()
	end

	-- check if it's time for the "automatic" event
	nextAutomaticTime = nextAutomaticTime - elapsedTime
	if nextAutomaticTime <= 0 then
		AttemptReaction("Automatic", 0, nil, playerName, nil, ActionType.HIT)
		Events:SetNextAutomaticTime()
	end

	-- check if you've still lost control of the character
	if nextLostControlCheck then
		nextLostControlCheck = nextLostControlCheck - elapsedTime
		if nextLostControlCheck <= 0 then
			nextLostControlCheck = 1
			AttemptReaction("LoseControl", 0, nil, playerName, nil, ActionType.HIT)
		end
	end

	-- check swimming/underwater - doesn't check if falling so that jumping won't stop swimming
	if not IsFalling() then
		nextSwimmingCheck = nextSwimmingCheck - elapsedTime
		if nextSwimmingCheck <= 0 then
			nextSwimmingCheck = 1
			if isUnderwater then
				AttemptReaction("Underwater", 0, nil, playerName, nil, ActionType.HIT)
			elseif IsSwimming() then
				if not isSwimming then
					isSwimming = true
					AttemptReaction("SwimmingBegin", 0, nil, playerName, nil, ActionType.HIT)
				else
					AttemptReaction("Swimming", 0, nil, playerName, nil, ActionType.HIT)
				end
			elseif isSwimming then
				isSwimming = nil
				AttemptReaction("SwimmingEnd", 0, nil, playerName, nil, ActionType.HIT)
			end
		end
	end

	-- check for a flight landing
	if nextFlightCheck then
		nextFlightCheck = nextFlightCheck - elapsedTime
		if nextFlightCheck <= 0 then
			if not flightVerified then -- not sure yet if the flight actually started, so check now
				if UnitOnTaxi("player") then
					flightVerified = true
					nextFlightCheck = 1
				else
					nextFlightCheck = nil
				end
			elseif UnitOnTaxi("player") then -- the flight has started and they're still on it
				nextFlightCheck = 1
			else -- the flight had started, but now they've landed
				flightVerified = nil
				nextFlightCheck = nil
				AttemptReaction("FlightMasterFlightEnd", 0, nil, playerName, flightLocationName, ActionType.HIT)
			end
		end
	end

	elapsedTime = 0 -- must be reset at the end so that the above checks can use it
end)

--------------------------------------------------
-- player loses or gains back control
--------------------------------------------------
function Events:PLAYER_CONTROL_GAINED()
	if UnitAffectingCombat("player") then
		AttemptReaction("LoseControlEnd", 0, nil, playerName, nil, ActionType.HIT)
	end
	nextLostControlCheck = nil
end

function Events:PLAYER_CONTROL_LOST()
	if UnitAffectingCombat("player") then
		AttemptReaction("LoseControlBegin", 0, nil, playerName, nil, ActionType.HIT)
		nextLostControlCheck = 1
	end
end

--------------------------------------------------
-- underwater checking
--------------------------------------------------
function Events:MIRROR_TIMER_START(timerType)
	if not isUnderwater and timerType == "BREATH" then
		isUnderwater = true
		isSwimming = true
		AttemptReaction("UnderwaterBegin", 0, nil, playerName, nil, ActionType.HIT)
		nextSwimmingCheck = 1
	end
end

function Events:MIRROR_TIMER_STOP(timerType)
	if isUnderwater and timerType == "BREATH" then
		isUnderwater = nil
		AttemptReaction("UnderwaterEnd", 0, nil, playerName, nil, ActionType.HIT)
		nextSwimmingCheck = 1
	end
end

--------------------------------------------------
-- flight master
--------------------------------------------------
function Events:TAXIMAP_OPENED()
	self.flightMasterOpened = UnitName("npc")
	AttemptReaction("FlightMasterOpen", 0, nil, playerName, self.flightMasterOpened, ActionType.HIT)
end

function Events:TAXIMAP_CLOSED()
	if self.flightMasterOpened and not nextFlightCheck then
		AttemptReaction("FlightMasterClose", 0, nil, playerName, self.flightMasterOpened, ActionType.HIT)
		self.flightMasterOpened = nil
	end
end

hooksecurefunc("TakeTaxiNode", function(index)
	local fullName = TaxiNodeName(index) or ""
	flightLocationName = fullName:match("([^,]+).*") or "parts unknown" -- saved so FlightMasterFlightLand can use it for <target_name>
	AttemptReaction("FlightMasterFlightBegin", 0, nil, playerName, flightLocationName, ActionType.HIT)
	nextFlightCheck = 3
end)