--[[
	Notes:
		* When both actions exist, "Group member dodges/resists" is used instead of "Non-grouped
		  person's spell misses/is resisted"

		* For channeling start/stop actions, <spell_link> only works if the spell is in the player's
		  spellbook because no ID is known. (need to check if that's different now)

	ReactionsSave:
		["enabled"]          - bool   - if automatic reactions are enabled
		["testing"]          - bool   - if messages will be printed instead of said/executed
		["globalCooldown"]   - number - wait this many seconds before using another reaction
		["messageCooldown"]  - number - individual message cooldown - avoid picking it again for this long unless impossible
		["chanceMultiplier"] - number - multiply the chance of reactions happening by this much
		["quietStealth"]     - bool   - if true, don't speak when stealthed and out in the world or a battleground
		["shoutMode"]        - bool   - if messages are converted to uppercase before using
		["roleItems"]        - table  - {[index] = {name, ID, texture, slot1, slot2}}
		["fightLength"]      - number - how long in seconds a fight must last to count as one
		["groupCounter"]     - number - counter for groups you join
		["fightCounter"]     - number - counter for fights started - only counts if they last at least 5 seconds
		["reactionList"]     - table  - {[name]=value,...} list of reactions - explained more below
		["groupList"]        - table  - {[name]=enabled,...} list of action groups and if they're enabled
		["tagList"]          - table  - {[name]={text="a|b|c",submenu=""}, ...} list of tags and their text
		["chatList"]         - table  - ReactionsSave.chatList["trigger"] = {[trigger name] = {settings}, ...}
			                                ReactionsSave.chatList["channel"] = {[channel name] = {"Trigger Name 1", "Trigger 2", ...}}

	ReactionsSave.reactionList["Spell"] =
		["Spell"].event    - bool/nil   - true if it's a special event instead of a spell
		["Spell"].submenu  - string/nil - Where it goes on the gui menu, like: "Druid>Feral>"
		["Spell"].nickname - string/nil - How it's shown on the menu (so you can name IDs with something recognizable)
		["Spell"].<action> - table      - hit/hit_by/miss/etc - table values explained below

	ReactionsSave.reactionList["Spell"]["<action>"] =
		["<action>"].group             - string/nil - the name of the group it's in
		["<action>"].chance            - number     - chance of reacting when it happens
		["<action>"].cooldown          - number/nil - overwrite global cooldown
		["<action>"].limitAura         - bool/nil   - only use it once until its aura is gone
		["<action>"].limitGroup        - bool/nil   - only use it once per group you're in
		["<action>"].limitFights       - bool/nil   - only use it once per limitFightsAmount of fights
		["<action>"].limitFightsAmount - number/nil - if limitFights is set, then how many fights must pass before able to use the reaction again
		["<action>"].limitName         - bool/nil   - only use it once on each target name per fight
		["<action>"].usedOnGroupNumber - number/nil - the ReactionsSave["groupCounter"] the reaction was used on if limitGroup is enabled
		["<action>"].usedOnFightNumber - number/nil - the ReactionsSave["fightCounter"] the reaction was used on if limitFights is enabled
		["<action>"].usedNames[onFight] = {[name]=1,[name]=1,...} - the names already used during this fight - table deleted/remade each new use to get rid of old names
		["<action>"].lastReactionTime  - number/nil - time() of last time the reaction was successful
		["<action>"].lastChosen        - number/nil - last reaction index number chosen
		["<action>"].<reactions>       - table      - {{table explained below}, ...}

	ReactionsSave.reactionList["Spell"]["action"][<reaction>] =
		[1] = channel
		[2] = action or message text
		[3] = last used time
		[4] = language to use (can be nil): nil or 1=common, 2=racial, 3=random
		[5] = table of various conditions (can be nil): if an item exists in the table then the reaction is NOT used if that condition is true (conditionsButtonMenu in gui files for list)
		[6] = role item: nil or 1 to 10 - for ReactionsSave.roleItems[number]
--]]

local _, RS = ...
RS.Reactions = {}
local Reactions = RS.Reactions

-- local references - WoW and Lua
local time                   = time
local GetTime                = GetTime
local UnitPlayerOrPetInParty = UnitPlayerOrPetInParty
local UnitPlayerOrPetInRaid  = UnitPlayerOrPetInRaid
local UnitAffectingCombat    = UnitAffectingCombat
local table_remove           = table.remove
local table_insert           = table.insert

-- local references - Addon
local Common        = nil
local Variables     = nil
local ActionType    = nil
local addonSettings = nil
local groupList     = nil
local trackedAuras  = nil
local playerName    = UnitName("player")
local ReactionConditionsUsable = nil
local ReactionWearingRoleItem  = nil

local lastReactionTime      = 0  -- the last time() a reaction was successfully used
local exactLastReactionTime = {} -- GetTime() of when a spell/event was last used for each action type

function Reactions:OnLoad()
	_G["GlobalTest"] = true
	Common        = RS.Common
	ActionType    = Common.ActionType
	Variables     = RS.Variables
	addonSettings = ReactionsSave
	groupList     = addonSettings.groupList
	reactionList  = addonSettings.reactionList
	trackedAuras  = RS.Events.trackedAuras
	ReactionConditionsUsable = RS.Conditions.ReactionConditionsUsable
	ReactionWearingRoleItem  = RS.Conditions.ReactionWearingRoleItem
	for _,v in pairs(ActionType) do
		exactLastReactionTime[v] = {}
	end
end

----------------------------------------------------------------------------------------------------
-- helper functions
----------------------------------------------------------------------------------------------------
-- find a spell/event in the reaction list - first by its ID then name
function Reactions.GetSpellSettings(spellName, spellId)
	return (spellId and reactionList[tostring(spellId)]) or (spellName and reactionList[spellName]) or nil
end
local GetSpellSettings = Reactions.GetSpellSettings

----------------------------------------------------------------------------------------------------
-- attempt to use a spell/event reaction - return true if successfully used
----------------------------------------------------------------------------------------------------
-- well this became a big mess over time!
-- spellName, spellId = the name and/or ID number of the spell or event to react to - no IDs in Classic
-- hitType = nil/Normal, Critical, Crushing, or Glancing - or miss type: MISS, DODGE, BLOCK, PARRY, etc
-- caster, target = the caster and target of the spell/event
-- actionType = value from the ActionType table (HIT/MISS/START_CAST/etc)
-- force = don't check certain things like cooldowns
-- forceMessage = just create a test message
-- redirectionSpellName = if using the "Spell/Event" channel, it will send this so that the <spellName> variable works
-- redirectionSpellId = if using the "Spell/Event" channel, it will send this so that the <spell_link> variable works
-- redirectionFromEvent = if the "Spell/Event" channel originated from an event - to be able to get you_hit/member_hit/other_hit properly
-- redirectionCount = counter of how many times "Spell/Event" channel has redirection to another reaction list - to be able to exit an endless loop
function Reactions.AttemptReaction(spellName, spellId, hitType, caster, target, actionType, force, forceMessage,
                                   redirectionSpellName, redirectionSpellId, redirectionFromEvent, redirectionCount, redirectionTime)
	if not addonSettings.enabled and not forceMessage then
		return
	end

	local spellSettings
	if not forceMessage then
		spellSettings = GetSpellSettings(spellName, spellId)
		if not spellSettings then
			return
		end

		-- only allow 1 trigger for each spell/event in a short period of time to stop those that have multiple events at once
		if not force then
			local lastUsed = exactLastReactionTime[actionType][spellId and spellId ~= 0 and spellId or spellName]
			if lastUsed and GetTime() - lastUsed < 0.7 and not redirectionCount then
				return
			end
		end
	end

	-- when someone casts on themselves, remove the target if there is one and and set a flag to know its a self-cast
	-- don't do this for events so that they can use a simpler system of just you_hit, member_hit, other_hit to tell who it's affecting
	local targetSelf = nil
	if spellSettings and not spellSettings.event and not redirectionFromEvent then
		if caster == target or target == nil then
			targetSelf = true
			target = nil
		end
	end

	-- figure out which message type to use and set up variable names
	local actionInfo       = nil -- refence to the action table that contains the settings and reactions
	local reactions        = nil -- reference to the list of reactions for the action
	local chosenReaction   = nil -- indice of the reaction that's chosen to be used
	local auraUnitName     = nil -- name of the person to check auras on in limitAura=true messages
	local memberName       = nil
	local targetName       = nil
	local extraTargetName  = nil
	local casterName       = caster

	local inGroup = IsInGroup()

	if forceMessage then
		casterName      = caster or playerName
		memberName      = target or "Illidan"
		targetName      = target or "Hogger"
		extraTargetName = target or "Wood Man"
	else
		-- figure out the type of action (you_hit, member_miss, etc)
		-- player casting something
		if (casterName == playerName or (casterName == nil and actionType ~= ActionType.EVENT_AURA)) and (targetSelf or (actionType ~= ActionType.AURA_GAINED and actionType ~= ActionType.AURA_REMOVED)) then
			actionInfo =  (actionType == ActionType.HIT and ((targetSelf and spellSettings.you_hit_self) or (not targetSelf and spellSettings.you_hit)))
			           or (actionType == ActionType.PERIODIC and ((targetSelf and spellSettings.periodic_you_hit_self) or (not targetSelf and spellSettings.periodic_you_hit)))
			           or (actionType == ActionType.AURA_GAINED and spellSettings.aura_gained_you)
			           or (actionType == ActionType.AURA_REMOVED and spellSettings.aura_removed_you)
			           or (actionType == ActionType.AURA_REMOVED and spellSettings.aura_removed_you)
						  or (actionType == ActionType.MISS and spellSettings.you_miss)
			           or (actionType == ActionType.START_CAST and spellSettings.start_cast)
			           or (actionType == ActionType.CHANNEL_STOP and spellSettings.you_channel_stop)
			           or nil -- end of assigning actionInfo
			targetName        = target
			memberName = (UnitPlayerOrPetInParty(target) or UnitPlayerOrPetInRaid(target) or UnitIsUnit(target or "player", "pet")) and target or nil
			auraUnitName = target or casterName
		-- someone casting something on the player
		elseif target == playerName then
			actionInfo =  (actionType == ActionType.HIT and spellSettings.you_get_hit)
			           or (actionType == ActionType.PERIODIC and spellSettings.periodic_you_get_hit)
			           or (actionType == ActionType.MISS and spellSettings.you_dodge)
			           or (actionType == ActionType.AURA_GAINED and spellSettings.aura_gained_you)
			           or (actionType == ActionType.AURA_REMOVED and spellSettings.aura_removed_you)
			           or (actionType == ActionType.EVENT_AURA and spellSettings.you_hit) -- spell effect events
			           or nil -- end of assigning actionInfo
			targetName        = casterName
			memberName = (UnitPlayerOrPetInParty(casterName) or UnitPlayerOrPetInRaid(casterName) or UnitIsUnit(casterName or "player", "pet")) and casterName or nil
			auraUnitName = target or casterName
		-- group member casting something
		elseif UnitPlayerOrPetInParty(casterName) or UnitPlayerOrPetInRaid(casterName) or UnitIsUnit(casterName or "player", "pet") then
			actionInfo =  (actionType == ActionType.HIT and ((targetSelf and spellSettings.member_hit_self) or (not targetSelf and spellSettings.member_hit)))
			           or (actionType == ActionType.PERIODIC and ((targetSelf and spellSettings.periodic_member_hit_self) or (not targetSelf and spellSettings.periodic_member_hit)))
			           or (actionType == ActionType.MISS and spellSettings.member_miss)
			           or (actionType == ActionType.START_CAST and spellSettings.member_start_cast)
			           or nil -- end of assigning actionInfo
			targetName        = target
			memberName = casterName
			auraUnitName = target or memberName
		-- someone casting something on a group member
		elseif UnitPlayerOrPetInParty(target) or UnitPlayerOrPetInRaid(target) or UnitIsUnit(target or "player", "pet") then
			actionInfo =  (actionType == ActionType.HIT and spellSettings.member_get_hit)
			           or (actionType == ActionType.PERIODIC and spellSettings.periodic_member_get_hit)
			           or (actionType == ActionType.MISS and spellSettings.member_dodge)
			           or (actionType == ActionType.AURA_GAINED and spellSettings.aura_gained_member)
			           or (actionType == ActionType.AURA_REMOVED and spellSettings.aura_removed_member)
			           or (actionType == ActionType.EVENT_AURA and spellSettings.member_hit) -- spell effect events
			           or nil -- end of assigning actionInfo
			targetName        = casterName or (actionType == ActionType.EVENT_AURA and target)
			memberName = target
			auraUnitName = target or memberName
			-- special case when there's no "Group member dodges/resists" but there is "Non-grouped person's spell misses/is resisted"
			if not actionInfo and actionType == ActionType.MISS and (not UnitPlayerOrPetInParty(caster) and not UnitPlayerOrPetInRaid(caster) and not UnitIsUnit(caster or "player", "pet")) then
				actionInfo = spellSettings.other_miss
			end
		-- non-group member casting or being hit by something
		else
			actionInfo =  (actionType == ActionType.HIT and ((targetSelf and spellSettings.other_hit_self) or (not targetSelf and spellSettings.other_hit)))
			           or (actionType == ActionType.PERIODIC and ((targetSelf and spellSettings.periodic_other_hit_self) or (not targetSelf and spellSettings.periodic_other_hit)))
			           or (actionType == ActionType.MISS and spellSettings.other_miss)
			           or (actionType == ActionType.START_CAST and spellSettings.other_start_cast)
			           or (actionType == ActionType.AURA_GAINED and spellSettings.aura_gained_other)
			           or (actionType == ActionType.AURA_REMOVED and spellSettings.aura_removed_other)
			           or (actionType == ActionType.EVENT_AURA and spellSettings.other_hit) -- spell effect events
			           or nil -- end of assigning actionInfo
			if actionType == ActionType.AURA_GAINED or actionType == ActionType.AURA_REMOVED or actionType == ActionType.EVENT_AURA then
				targetName = target
			else
				targetName = casterName
				extraTargetName = target
			end
			auraUnitName = extraTargetName or targetName
		end

		if not actionInfo then
			return
		end

		if not force then
			-- set the chance modifier and check the chance
			local chance = actionInfo.chance
			if not chance or chance * addonSettings.chanceMultiplier < random() * 100 then
				return
			end
			-- check if it's in a group that's disabled
			if (actionInfo.group and not groupList[actionInfo.group]) or (not actionInfo.group and not groupList["Ungrouped"]) then
				return
			end
			-- if it has a travel time, then it must have a hit type or else UNIT_SPELLCAST_SUCCEEDED would trigger it as soon as the cast was complete
			if not hitType and spellSettings.travelTime and actionType ~= ActionType.START_CAST then
				return
			end
			-- check cooldown time
			if actionInfo.cooldown then
				if actionInfo.lastReactionTime and actionInfo.lastReactionTime > time() - actionInfo.cooldown then
					return
				end
			elseif lastReactionTime > time() - addonSettings.globalCooldown then
				return
			end
			-- check limitAura
			if actionInfo.limitAura and (UnitPlayerOrPetInRaid(auraUnitName) or UnitPlayerOrPetInParty(auraUnitName)) then
				local auraName = spellName or GetSpellInfo(spellId)
				if trackedAuras[auraUnitName] and trackedAuras[auraUnitName][auraName] then
					return
				end
				if not Common.HasAura(auraUnitName, auraName) then
					return
				end
			end
			-- check limitGroup - if they're in the same group as the last time it was used then stop
			if actionInfo.limitGroup and inGroup and actionInfo.usedOnGroupNumber and actionInfo.usedOnGroupNumber == addonSettings.groupCounter then
				return
			end
			-- check limitFights - if enough fights haven't passed since the last time it was used then stop
			if actionInfo.limitFights and actionInfo.usedOnFightNumber and addonSettings.fightCounter < actionInfo.usedOnFightNumber + actionInfo.limitFightsAmount then
				return
			end
			-- check limitName - if a name has been involved with this action during this fight, then stop
			if actionInfo.limitName then
				local list = actionInfo.usedNames and actionInfo.usedNames[addonSettings.fightCounter]
				if list and auraUnitName and list[auraUnitName] then
					return
				end
			end
		end

		-- check if any reactions exist to use
		reactions = actionInfo.reactions
		if not reactions or next(reactions) == nil then
			return
		end

		local lastChosen = actionInfo.lastChosen
		local totalReactions = #reactions

		-- create a list of possible things to say - save the indice from the reactions table so the last chosen and time used can be updated if needed.
		-- reactions[i][3] is the last time used, [4] is language used, [5] is table of not allowed conditions
		local possibleReactions = {}
		local possibleCount = 0

		-- set combat and form variables to be able to compare them easier with each reaction
		local usableCombat = UnitAffectingCombat("player") and "InCombat" or "NoCombat"
		local currentForm
		local formIndex = GetShapeshiftForm()
		if formIndex == 0 then
			currentForm = "Normal Form"
		else
			currentForm = tostring((select(4, GetShapeshiftFormInfo(formIndex))))
			if currentForm == "9634" then -- Dire Bear Form to Bear Form
				currentForm = "5487"
			end
		end

		Variables.targetName = targetName
		Variables.casterName = casterName
		Variables.memberName = memberName
		Variables.extraTargetName = extraTargetName

		-- clear the cache that keeps the more intensive checking results stored to reuse through all
		-- the checking here
		conditionFriendCache = {}
		conditionGuildCache = {}

		-- first add ones that have not been used yet or in a long enough time
		-- TODO: rethink the whole random picking process - shuffling would normally be better but
		--       it's not that simple now that there are restrictions on using each reaction
		for i=1,totalReactions do
			if i ~= lastChosen
			 and (not reactions[i][3] or reactions[i][3] < time()-addonSettings.messageCooldown)
			 and (reactions[i][1] ~= "Group" or inGroup)
			 and ReactionConditionsUsable(reactions[i][5], usableCombat, currentForm, hitType, spellId)
			 and (not reactions[i][6] or ReactionWearingRoleItem(reactions[i][6])) then
				table_insert(possibleReactions, i)
				possibleCount = possibleCount + 1
			end
		end

		-- if none were found, add some messages that are on cooldown
		if possibleCount == 0 then
			-- first create a list of the messages on cooldown so they can be picked among equally
			local cooldowns = {}
			for i=1,totalReactions do
				if i ~= lastChosen
				 and (reactions[i][3] and reactions[i][3] >= time()-addonSettings.messageCooldown)
				 and (reactions[i][1] ~= "Group" or inGroup)
				 and ReactionConditionsUsable(reactions[i][5], usableCombat, currentForm, hitType, spellId)
				 and (not reactions[i][6] or ReactionWearingRoleItem(reactions[i][6])) then
					table_insert(cooldowns, i)
				end
			end
			-- keep picking until there's enough in the possible message list
			local cooldownCount = #cooldowns
			if cooldownCount > 0 then
				local needed = 3
				repeat
					local picked = random(1, cooldownCount)
					table_insert(possibleReactions, cooldowns[picked])
					table_remove(cooldowns, picked)
					cooldownCount = cooldownCount - 1
					possibleCount = possibleCount + 1
					needed = needed - 1
				until needed <= 0 or cooldownCount <= 0
			end
		end

		-- if nothing was possible, then add the last chosen one if it exists and is possible to use
		if possibleCount == 0
		 and lastChosen
		 and (reactions[lastChosen][1] ~= "Group" or inGroup)
		 and ReactionConditionsUsable(reactions[lastChosen][5], usableCombat, currentForm, hitType, spellId)
		 and (not reactions[lastChosen][6] or ReactionWearingRoleItem(reactions[lastChosen][6])) then
			possibleReactions[1] = lastChosen
			possibleCount = 1
		end

		if possibleCount <= 0 then
			return
		end
		chosenReaction = possibleReactions[random(1, possibleCount)]
	end

	-- update cooldowns, limits, and last chosen
	if not force then
		actionInfo.lastChosen = chosenReaction
		actionInfo.lastReactionTime = time()
		reactions[chosenReaction][3] = time()
		if not actionInfo.noGCD then
			lastReactionTime = time()
		end
		exactLastReactionTime[actionType][spellId and spellId ~= 0 and spellId or spellName] = GetTime()

		-- limitGroup
		if actionInfo.limitGroup and inGroup then
			actionInfo.usedOnGroupNumber = addonSettings.groupCounter
		end
		-- limitFights
		if actionInfo.limitFights then
			actionInfo.usedOnFightNumber = addonSettings.fightCounter
		end
		-- limitAura
		if actionInfo.limitAura and (UnitPlayerOrPetInParty(auraUnitName) or UnitPlayerOrPetInRaid(auraUnitName)) then
			local name = spellName or GetSpellInfo(spellId)
			trackedAuras[auraUnitName] = trackedAuras[auraUnitName] or {}
			trackedAuras[auraUnitName][name] = true
		end
		--limitName
		if actionInfo.limitName and auraUnitName then
			local list = actionInfo.usedNames and actionInfo.usedNames[addonSettings.fightCounter]
			if not list then
				actionInfo.usedNames = {} -- rebuild it to erase any old tables
				actionInfo.usedNames[addonSettings.fightCounter] = {}
				list = actionInfo.usedNames[addonSettings.fightCounter]
			end
			list[auraUnitName] = true
		end
	end

	local Variables = Variables
	-- some events (like Interrupting) may set current spells, so don't overwrite those
	if spellSettings and not spellSettings.event and not redirectionFromEvent then
		Variables.spellName = redirectionSpellName or spellName
		Variables.spellId = redirectionSpellId or spellId
	end

	if forceMessage then
		RS.Messages.ParseAndSendMessage(forceMessage, "force_message")
	elseif chosenReaction then
		local reactionData = {
			hitType,
			casterName,
			targetName,
			actionType,
			force,
			forceMessage,
			redirectionSpellName or spellName,
			redirectionSpellId or spellId,
			redirectionFromEvent or (spellSettings and spellSettings.event),
			redirectionCount
		}
		RS.Messages.ParseAndSendMessage(reactions[chosenReaction][2], reactions[chosenReaction][1], targetName,
		                                reactions[chosenReaction][4], reactionData, redirectionTime)
	end
	return true
end
